# Copyright 2025 John Sirois.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import annotations

import base64
import hashlib
import importlib.util
import json
import os
import shutil
import subprocess
import sys
from contextlib import contextmanager
from dataclasses import dataclass
from os import fspath
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from textwrap import dedent
from typing import IO, Any, Dict, Iterator, cast

from dev_cmd import color
from dev_cmd.model import Command, PythonConfig, Venv

AVAILABLE = False

if shutil.which("pex3") and importlib.util.find_spec("filelock"):
    from filelock import FileLock

    AVAILABLE = True


def _fingerprint(data: bytes) -> str:
    return base64.urlsafe_b64encode(hashlib.sha256(data).digest()).decode()


@contextmanager
def named_temporary_file(
    tmp_dir: str | None = None, prefix: str | None = None
) -> Iterator[IO[bytes]]:
    # Work around Windows issue with auto-delete: https://bugs.python.org/issue14243
    fp = NamedTemporaryFile(dir=tmp_dir, prefix=prefix, delete=False)
    try:
        with fp:
            yield fp
    finally:
        try:
            os.remove(fp.name)
        except FileNotFoundError:
            pass


def _ensure_cache_dir() -> Path:
    cache_dir = Path(os.path.abspath(os.environ.get("DEV_CMD_WORKSPACE_CACHE_DIR", ".dev-cmd")))
    gitignore = cache_dir / ".gitignore"
    if not gitignore.exists():
        cache_dir.mkdir(parents=True, exist_ok=True)
        with named_temporary_file(tmp_dir=fspath(cache_dir), prefix=".gitignore.") as gitignore_fp:
            gitignore_fp.write(b"*\n")
            gitignore_fp.close()
            os.rename(gitignore_fp.name, gitignore)
    return cache_dir


@dataclass(frozen=True)
class VenvLayout:
    python: str
    site_packages_dir: str


def _create_venv(python: str, venv_dir: str) -> VenvLayout:
    subprocess.run(
        args=[
            "pex3",
            "venv",
            "create",
            "--force",
            "--python",
            python,
            "--pip",
            "--dest-dir",
            venv_dir,
        ],
        check=True,
    )

    result = subprocess.run(
        args=["pex3", "venv", "inspect", venv_dir],
        text=True,
        stdout=subprocess.PIPE,
        check=True,
    )
    venv_data = json.loads(result.stdout)
    python_exe = venv_data["interpreter"]["binary"]
    site_packages_dir = venv_data["site_packages"]

    return VenvLayout(python=python_exe, site_packages_dir=site_packages_dir)


def marker_environment(python: str) -> dict[str, str]:
    fingerprint = _fingerprint(python.encode())
    markers_file = _ensure_cache_dir() / "interpreters" / f"markers.{fingerprint}.json"
    if not os.path.exists(markers_file):
        with FileLock(f"{markers_file}.lck"), TemporaryDirectory(
            dir=markers_file.parent, prefix="packaging-venv."
        ) as td:
            print(
                f"{color.yellow(f'Calculating environment markers for --python {python}')}...",
                file=sys.stderr,
            )
            venv_layout = _create_venv(python, fspath(td))
            subprocess.run(
                args=[venv_layout.python, "-m", "pip", "install", "packaging"],
                stdout=sys.stderr.fileno(),
                check=True,
            )
            temp_markers_file = Path(td) / markers_file.name
            temp_markers_file.write_bytes(
                subprocess.run(
                    args=[
                        venv_layout.python,
                        "-c",
                        dedent(
                            """\
                            import json
                            import sys

                            from packaging import markers

                            json.dump(markers.default_environment(), sys.stdout)
                            """
                        ),
                    ],
                    stdout=subprocess.PIPE,
                    check=True,
                ).stdout
            )
            temp_markers_file.rename(markers_file)
    return cast(Dict[str, str], json.loads(markers_file.read_bytes()))


def _fingerprint_python_config(python: str, config: PythonConfig) -> str:
    input_files = {}
    input_paths: dict[str, str] = {}
    for path in config.cache_key_inputs.paths:
        input_path = Path(path)
        if input_path.is_file():
            input_files[input_path] = ""
        else:
            assert input_path.is_dir(), (
                f"Expected parsing code to verify {input_path} was an existing file or directory."
            )
            for root, dirs, files in os.walk(input_path):
                root_dir = Path(root)
                input_paths.update(((root_dir / d).as_posix(), "") for d in dirs)
                input_files.update((root_dir / f, "") for f in files)

    if input_files:
        input_paths.update(
            zip(
                map(lambda f: f.as_posix(), input_files),
                # TODO(John Sirois): Investigate ThreadPool().map(..., chunk_size=?) ~# of files
                #  threshold where performance improves A quick experiment on SSD showed a thread
                #  pool to be slower for <100 files.
                map(lambda f: _fingerprint(f.read_bytes()), input_files),
            )
        )

    def extract_command_fingerprint_data(command: Command | None) -> dict[str, Any] | None:
        if command is None:
            return None
        return {
            "extra-env": dict(command.extra_env),
            "args": command.args,
            "cwd": str(command.cwd) if command.cwd else None,
        }

    return _fingerprint(
        json.dumps(
            {
                "python": python,
                "cache-keys": {
                    "pyproject-data": config.cache_key_inputs.pyproject_data,
                    "paths": input_paths,
                    "env": config.cache_key_inputs.envs,
                },
                "pip-requirement": config.pip_requirement,
                "3rdparty-export-command": extract_command_fingerprint_data(
                    config.thirdparty_export_command
                ),
                "3rdparty-pip-install-opts": config.thirdparty_pip_install_opts,
                "extra-requirements": (
                    _fingerprint(config.extra_requirements.encode())
                    if isinstance(config.extra_requirements, str)
                    else config.extra_requirements
                ),
                "extra-requirements-pip-install-opts": config.extra_requirements_pip_install_opts,
                "finalize-command": extract_command_fingerprint_data(config.finalize_command),
            },
            sort_keys=True,
        ).encode()
    )


def ensure(python: str, config: PythonConfig, rebuild_if_needed: bool = True) -> Venv:
    fingerprint = _fingerprint_python_config(python=python, config=config)
    venv_dir = _ensure_cache_dir() / "venvs" / fingerprint
    layout_file = venv_dir / ".dev-cmd-venv-layout.json"
    if not os.path.exists(venv_dir):
        with FileLock(f"{venv_dir}.lck"):
            if not os.path.exists(venv_dir):
                print(
                    f"{color.yellow(f'Setting up venv for --python {python}')}...", file=sys.stderr
                )
                work_dir = Path(f"{venv_dir}.work")
                venv_layout = _create_venv(python, venv_dir=fspath(work_dir))
                with named_temporary_file(
                    tmp_dir=fspath(work_dir), prefix="3rdparty-reqs."
                ) as reqs_fp:
                    reqs_fp.close()

                    requirements_export_command_args = [
                        (reqs_fp.name if arg == "{requirements.txt}" else arg)
                        for arg in config.thirdparty_export_command.args
                    ]
                    env = os.environ.copy()
                    env.update(config.thirdparty_export_command.extra_env)
                    subprocess.run(
                        args=requirements_export_command_args,
                        cwd=config.thirdparty_export_command.cwd,
                        env=env,
                        check=True,
                    )

                    subprocess.run(
                        args=[
                            venv_layout.python,
                            "-m",
                            "pip",
                            "install",
                            "-U",
                            config.pip_requirement,
                        ],
                        stdout=sys.stderr.fileno(),
                        check=True,
                    )
                    subprocess.run(
                        args=[venv_layout.python, "-m", "pip", "install"]
                        + list(config.thirdparty_pip_install_opts)
                        + ["-r", reqs_fp.name],
                        stdout=sys.stderr.fileno(),
                        check=True,
                    )

                if config.extra_requirements:

                    @contextmanager
                    def _extra_requirements_args() -> Iterator[list[str]]:
                        if isinstance(config.extra_requirements, str):
                            with named_temporary_file(
                                tmp_dir=fspath(work_dir), prefix="extra-reqs."
                            ) as fp:
                                fp.write(config.extra_requirements.encode())
                                fp.close()
                                yield ["-r", fp.name]
                        else:
                            yield list(config.extra_requirements)

                    with _extra_requirements_args() as extra_requirements_args:
                        subprocess.run(
                            args=[venv_layout.python, "-m", "pip", "install"]
                            + list(config.extra_requirements_pip_install_opts)
                            + extra_requirements_args,
                            stdout=sys.stderr.fileno(),
                            check=True,
                        )

                if config.finalize_command:
                    finalize_command_args: list[str] = []
                    for arg in config.finalize_command.args:
                        if arg == "{venv-python}":
                            finalize_command_args.append(venv_layout.python)
                        elif arg == "{venv-site-packages}":
                            finalize_command_args.append(venv_layout.site_packages_dir)
                        else:
                            finalize_command_args.append(arg)
                    env = os.environ.copy()
                    env.update(config.finalize_command.extra_env)
                    subprocess.run(
                        args=finalize_command_args,
                        cwd=config.finalize_command.cwd,
                        env=env,
                        check=True,
                    )

                with (work_dir / layout_file.name).open("w") as out_fp:
                    json.dump(
                        {
                            "python": venv_layout.python.replace(str(work_dir), str(venv_dir)),
                            "marker-environment": marker_environment(python),
                        },
                        out_fp,
                    )
                work_dir.rename(venv_dir)

    with layout_file.open() as in_fp:
        data = json.load(in_fp)

    print(
        color.color(f"Using venv at {venv_dir} for --python {python}.", fg="gray"), file=sys.stderr
    )
    try:
        return Venv(
            dir=venv_dir.as_posix(),
            python=data["python"],
            marker_environment=data["marker-environment"],
        )
    except KeyError:
        if not rebuild_if_needed:
            raise
        print(
            color.yellow(f"Venv for --python {python} at {venv_dir} is out of date, rebuilding."),
            file=sys.stderr,
        )
        shutil.rmtree(venv_dir)
        return ensure(python=python, config=config, rebuild_if_needed=False)
